import time
import math
import numpy as np
import dynamico.wrap as wrap
from ctypes import c_void_p, c_int, c_double, c_bool

#------------- direct Cython interface to DYNAMICO routines -------------#

cdef extern from "functions.h":
     cdef void dynamico_init_params()
     cpdef void dynamico_setup_xios()
     cpdef void dynamico_xios_set_timestep(double)
     cpdef void dynamico_xios_update_calendar(int)

#------------- import and wrap DYNAMICO routines -------------#

ker=wrap.Struct() # store imported fun X as funs.X

check_args = False # use True instead of False for debugging, probably with some overhead

try:    
    kernels = wrap.SharedLib(vars(ker), 'libicosa.so', check_args=check_args) 
    setvar, setvars, getvar, getvars = kernels.setvar, kernels.setvars, kernels.getvar, kernels.getvars
except OSError:
    print """
Unable to load shared library 'libkernels.so' !
    """
    raise

# providing a full prototype enables type-checking when calling
# if a number n is present in the prototype, the previous type is repeated n times
kernels.import_funs([
                     ['dynamico_setup_xios',None],
                     ['dynamico_xios_set_timestep',c_double],
                     ['dynamico_xios_update_calendar',c_int],
                     ['dynamico_init_mesh',c_void_p,13],
                     ['dynamico_init_metric', c_void_p,6],
                     ['dynamico_init_hybrid', c_void_p,3],
                     ['dynamico_caldyn_unstructured', c_double, c_void_p,20],
                     ['dynamico_partition_graph', c_int,2, c_void_p,3, c_int, c_void_p],
                     ])

# set/get global variables
eta_mass,eta_lag=(1,2)
thermo_theta,thermo_entropy,thermo_moist,thermo_boussinesq=(1,2,3,4)

kernels.addvars(
    c_bool,'hydrostatic','debug_hevi_solver','rigid',
    c_int,'llm','nqdyn','primal_num','max_primal_deg',
    'dual_num','max_dual_deg','edge_num','max_trisk_deg',
    'caldyn_thermo','caldyn_eta','nb_threads',
    c_double,'elapsed','g', 'ptop', 'cpp', 'cppv',
    'Rd', 'Rv', 'preff', 'Treff', 'pbot', 'rho_bot', 'Phi_bot')

elapsed=0.

#----------------------------- Base class for dynamics ------------------------

class Caldyn:
    def __init__(self,mesh):
        self.mesh=mesh
        fps, ftheta, fmass = mesh.field_ps, mesh.field_theta, mesh.field_mass
        fw, fu, fz = mesh.field_w, mesh.field_u, mesh.field_z
        self.ps, self.ms, self.dms       = fps(), fps(), fps()
        self.s, self.hs, self.dhs        = ftheta(), ftheta(), ftheta()
        self.pk, self.berni, self.geopot, self.hflux = fmass(),fmass(),fw(),fu()
        self.qu, self.qv                 = fu(),fz()
        self.fmass, self.ftheta, self.fu, self.fw = fmass, ftheta, fu, fw
    def bwd_fast_slow(self, flow, tau):
        global elapsed
        time1=time.time()
        flow,fast,slow = self._bwd_fast_slow_(flow,tau)
        time2=time.time()
        elapsed=elapsed+time2-time1
        return flow,fast,slow

# when calling caldyn_unstructured, arrays for tendencies must be re-created each time
# to avoid overwriting in the same memory space when time scheme is multi-stage

#-------------------------- Shallow-water dynamics ---------------------

class Caldyn_RSW(Caldyn):
    def __init__(self,mesh):
        Caldyn.__init__(self,mesh)
        setvars(('hydrostatic','caldyn_thermo','caldyn_eta'),
                (True,thermo_boussinesq,eta_lag))
        self.dhs = self.fmass()
        dynamico_init_params()
    def _bwd_fast_slow_(self, flow, tau):
        h,u = flow
        # h*s = h => uniform buoyancy s=1 => shallow-water
        dh, du_slow, du_fast, hs, buf = self.fmass(), self.fu(), self.fu(), h.copy(), self.geopot
        ker.dynamico_caldyn_unstructured(tau, self.ms, h, hs, u, self.geopot, buf,
                  self.s, self.ps, self.pk, self.hflux, self.qv,
                  self.dms, dh, self.dhs, du_fast, du_slow,
                  buf, buf, buf, buf)
        return (h,u), (0.,du_fast), (dh,du_slow)

#----------------------------------- HPE ------------------------------------

class Caldyn_HPE(Caldyn):
    def __init__(self,caldyn_thermo,caldyn_eta, mesh,metric,thermo,BC,g):
        Caldyn.__init__(self,mesh)
        setvars(('hydrostatic','caldyn_thermo','caldyn_eta',
                 'g','ptop','Rd','cpp','preff','Treff'),
                (True,caldyn_thermo,caldyn_eta,
                 g,BC.ptop,thermo.Rd,thermo.Cpd,thermo.p0,thermo.T0))
        dynamico_init_params()
    def _bwd_fast_slow_(self, flow, tau):
        dm, dS, du_slow, du_fast, buf = self.fmass(), self.ftheta(), self.fu(), self.fu(), self.geopot
        m,S,u = flow
        ker.dynamico_caldyn_unstructured(tau, self.ms, m, S, u, self.geopot, buf,
                  self.s, self.ps, self.pk, self.hflux, self.qv,
                  self.dms, dm, dS, du_fast, du_slow,
                  buf, buf, buf, buf)
        return (m,S,u), (0.,0.,du_fast), (dm,dS,du_slow)

#----------------------------------- NH ------------------------------------

class Caldyn_NH(Caldyn):
    def __init__(self,caldyn_thermo,caldyn_eta, mesh,metric,thermo,BC,g):
        Caldyn.__init__(self,mesh)
        setvars(('hydrostatic','caldyn_thermo','caldyn_eta',
                 'g','ptop','Rd','cpp','preff','Treff',
                 'pbot','rho_bot'),
                (False,caldyn_thermo,caldyn_eta,
                 g,BC.ptop,thermo.Rd,thermo.Cpd,thermo.p0,thermo.T0,
                 BC.pbot.max(), BC.rho_bot.max()))
        dynamico_init_params()
    def bwd_fast_slow(self, flow, tau):
        ftheta, fmass, fu, fw = self.ftheta, self.fmass, self.fu, self.fw 
        dm, dS, du_slow, du_fast = fmass(), ftheta(), fu(), fu()
        dPhi_slow, dPhi_fast, dW_slow, dW_fast = fw(), fw(), fw(), fw()
        m,S,u,Phi,W = flow
        ker.dynamico_caldyn_unstructured(tau, self.ms, m, S, u, Phi, W,
                  self.s, self.ps, self.pk, self.hflux, self.qv,
                  self.dms, dm, dS, du_fast, du_slow,
                  dPhi_fast, dPhi_slow, dW_fast, dW_slow)
        return ((m,S,u,Phi,W), (0.,0.,du_fast,dPhi_fast,dW_fast), 
                (dm,dS,du_slow,dPhi_slow,dW_slow))

#------------------------ Copy mesh info to Fortran side -------------------

def init_mesh(llm, nqdyn, edge_num, primal_num, dual_num,
              max_trisk_deg, max_primal_deg, max_dual_deg,
              primal_nb, primal_edge, primal_ne,
              dual_nb,dual_edge,dual_ne,dual_vertex, 
              left,right,down,up,trisk_deg,trisk,
              Ai, Av, fv, le_de, Riv2, wee):
    setvars( ('llm','nqdyn','edge_num','primal_num','dual_num',
              'max_trisk_deg','max_primal_deg','max_dual_deg'),
             (llm, nqdyn, edge_num, primal_num, dual_num, 
              max_trisk_deg, max_primal_deg, max_dual_deg) )        
    print('init_mesh ...')
    ker.dynamico_init_mesh(primal_nb,primal_edge,primal_ne,
              dual_nb,dual_edge,dual_ne,dual_vertex,
              left,right,down,up,trisk_deg,trisk)
    print ('...done')
    print('init_metric ...')
    ker.dynamico_init_metric(Ai,Av,fv,le_de,Riv2,wee)
    print ('...done')

#------------------------ Mesh partitioning ------------------------

# Helper functions and interface to ParMETIS
# list_stencil converts an adjacency graph from array format index[num_cells, MAX_EDGES] to compressed format
# loc_stencil returns the start/end indices (vtxdist) expected by ParMETIS
# i.e. index[start:end] with start=vtxdist[cell], end=vtxdist[cell+1] lists the edges of cell 'cell'

def list_stencil(degree, stencil, cond=lambda x:True):
    for i in range(degree.size):
        for j in range(degree[i]):
            s=stencil[i,j]
            if cond(s): yield stencil[i,j]
                
def loc_stencil(degree, stencil):
    loc=0
    for i in range(degree.size):
        yield loc
        loc=loc+degree[i]
    yield loc

def partition_mesh(degree, stencil, nparts): 
    # arguments : PArray1D and PArray2D describing mesh, number of desired partitions
    dim_cell, degree, stencil = degree.dim, degree.data, stencil.data
    comm, vtxdist, idx_start, idx_end = dim_cell.comm, dim_cell.vtxdist, dim_cell.start, dim_cell.end
    mpi_rank, mpi_size = comm.Get_rank(), comm.Get_size()
    adjncy_loc, xadj_loc = list_stencil(degree, stencil), loc_stencil(degree, stencil)
    adjncy_loc, xadj_loc = [np.asarray(list(x), dtype=np.int32) for x in (adjncy_loc, xadj_loc)]
    owner = np.zeros(idx_end-idx_start, dtype=np.int32);
    ker.dynamico_partition_graph(mpi_rank, mpi_size, vtxdist, xadj_loc, adjncy_loc, nparts, owner)
    return owner
